import click
from stairs.core.session import project_session


@click.group()
def consumer_cli():
    pass


@consumer_cli.command("consumer:standalone")
@click.argument('name')
def init_session(name):
    project = project_session.get_project()
    print("init standalone consumer session")
    app_name, consumer_name = name.split('.')
    user_app = project.get_app_by_name(app_name)
    user_app.components.consumers[consumer_name].run_worker()




from .app_config import app

from .flows import FlowSample
from .consumers import print_smth


@app.pipeline()
def main(pipeline, data_frame):
    return data_frame\
        .make(data=data_frame)\
        .subscribe_flow(FlowSample)\
        .subscribe_consumer(print_smth)






import string
import json


from stairs.core.flow import Flow, step
from threading import Thread


class NameExtractionOneWayFlow(Flow):

    def __init__(self, use_lower=False):
        self.use_lower = use_lower

    def __call__(self, sentence):
        print("Im in NameExtractionOneWayFlow ")
        r = self.start_from(self.clean,
                            sentence=sentence)
        print(r)
        return r.result_name

    @step(None)
    def result_name(self, words):
        print("get name")
        result_names = []
        for name in ['Oleg']:
            if name in words:
                if self.use_lower:
                    name = name.lower()
                result_names.append(name)

        return dict(names=result_names)

    @step(result_name)
    def split_by_words(self, sentence):
        print("Split by words")
        return dict(words=sentence.split())

    @step(split_by_words)
    def remove_punctuation(self, sentence):
        print("remove_punctuation")
        for p in string.punctuation:
            sentence = sentence.replace(p, ' ')

        return dict(sentence=sentence)

    @step(remove_punctuation)
    def clean(self, sentence):
        print("clean")
        sentence = sentence.strip()
        return dict(sentence=sentence)


class NameExtractionFlowMultiple(NameExtractionOneWayFlow):

    def __call__(self, sentence):
        print("Im in NameExtractionFlowMultiple ")
        r = self.start_from(self.clean,
                            sentence=sentence)
        print(r)
        return dict(
            names=r.result_name['names'],
            sentence=r.clean['sentence']
        )

    @step(NameExtractionOneWayFlow.remove_punctuation, save_result=True)
    def clean(self, sentence):
        print("clean")
        sentence = sentence.strip()
        return dict(sentence=sentence)


class NameExtractionWorkersOneWayFlow(NameExtractionOneWayFlow):

    def __reconnect__(self):
        self.result_name.set_next(self.save_worker_result)

    def __call__(self, sentence):
        result = self.start_from(self.clean,
                                 sentence=sentence)

        return dict()

    @step(None)
    def save_worker_result(self, **kwargs):
        self.redis.set("test", json.dumps(kwargs))




import click
import stepist

from stairs.core import app as apps_worker
from stairs.core.flow.step import StairsStepAbstract

from stairs.core.session import project_session

from stepist.flow.session import get_steps_to_listen, get_step_by_key
from stairs.core.output.standalone import StandAloneConsumer


@click.group()
def workers_cli():
    pass


@workers_cli.command("pipelines:run")
@click.argument('apps', nargs=-1)
def run(apps):
    """
    Run workers process
    """

    project = project_session.get_project()
    project.run_pipelines()



from typing import List
from stairs.core.utils import AttrDict
from stepist.app_config import AppConfig as StepistConfig


class ProjectConfig(StepistConfig):
    # list of apps modules path

    @classmethod
    def init_default(cls):
        stepist_config = StepistConfig.init_default()
        return stepist_config

    @classmethod
    def load_from_file(cls, filename):
        """Load a config from a python file."""

        if isinstance(filename, str):
            d = dict()

            with open(filename) as f:
                code = compile(f.read(), filename, 'exec')
                exec(code, d)
        else:
            d = vars(filename)

        d = AttrDict(d)
        return d



import inspect

from stairs.core.app import components


class Output(components.AppOutput):

    def __init__(self, app, handler, as_daemon=False):
        self.handler = handler
        self.as_daemon = as_daemon
        self.app = app

        self.step = self.app\
            .project\
            .stepist_app\
            .step(None, as_worker=as_daemon)(self)

        components.AppOutput.__init__(self, self.app)

    def __call__(self, **data):
        output_result = self.handler(**data)

        if not self.as_daemon:
            return output_result

    def __name__(self):
        module_name = inspect.getmodule(self.handler).__name__
        return "%s:%s" % (self.handler.__name__, module_name)

    def step_key(self):
        return self.__name__()

    def key(self):
        return self.name()

    def name(self):
        return self.handler.__name__

    def get_stepist_step(self):
        return self.step





#!/usr/bin/env python
from stairs.services.management import init_cli
from stairs.core.project import StairsProject

if __name__ == "__main__":
    stairs_project = StairsProject(config_file="config.py")
    init_cli()



import uuid

from stairs.core.utils import AttrDict

from stepist.flow.utils import validate_handler_data
from stepist.flow.steps.hub import Hub
from stepist.flow.steps.next_step import call_next_step

from stairs.core.app.components import AppStep


def step(*next_steps, save_result=False, name=None):
    def wrapper(handler):
        s = FlowStep(handler, next_steps, save_result=save_result, name=name)
        handler.__meta__ = dict(step=s)
        return handler
    return wrapper


class StairsStepAbstract:
    # handler function which will be process data
    handler = None

    flow = None

    # stepist.step object
    step = None

    id = None

    def __name__(self):
        # All steps inside flow are not workers.
        # Let's add uuid4 for prevent instance names duplication for steps.
        return "%s:%s:%s" % (self.flow.key(),
                             self.handler.__name__,
                             self.id)

    def set_next(self, *next_steps):
        if len(next_steps) == 1:
            self.step.next_step = next_steps[0].step
        else:
            self.step.next_step = Hub(*[s.step for s in next_steps])

    def add_next(self, *next_steps):
        if self.step.next_step is not None:
            self.step.next_step = Hub(*[s.step for s in next_steps] +
                                       [self.step.next.step])
        else:
            self.set_next(*next_steps)

    def set(self, save_result=None):
        if save_result is not None:
            self.step.save_result = True

    def key(self):
        return self.__name__()

    def get_stepist_step(self):
        return self.step


class StairsStep(StairsStepAbstract, AppStep):
    step = None

    def __init__(self, pipeline, flow, handler, next_steps, save_result, name):
        self.flow = flow

        self.pipeline = pipeline
        self.handler = handler
        self.next_steps = next_steps

        self.name = name or self.handler.__name__

        self.id = uuid.uuid4()

        if len(next_steps) == 1:
            stepist_next_step = next_steps[0]
            if stepist_next_step is not None:
                stepist_next_step = stepist_next_step.step
        else:
            stepist_next_step = Hub(*[s.step for s in next_steps])

        self.step = self.pipeline\
            .app\
            .project\
            .stepist_app\
            .step(stepist_next_step,
                  unique_id=self.key(),
                  save_result=save_result,
                  name=self.name
            )(self.execute_step)

        AppStep.__init__(self, self.pipeline.app)

    def __call__(self, **data):
        flow_result = call_next_step(data, self.step)

        return flow_result

    def execute_step(self, **data):
        handler_data = validate_handler_data(self.handler, data)

        new_data = self.handler(self.flow, **handler_data)

        if new_data:
            data.update(new_data)

        if self.step.next_step is None:
            return new_data

        return data


class FlowStep:
    def __init__(self, handler, next_steps, name=None, save_result=False):
        self.handler = handler
        self.next_steps = next_steps
        self.save_result = save_result
        self.name = name





import importlib

from stairs.core.utils import AttrDict

from stairs.core.app import components
from stairs.core.app.components_interface import ComponentsMixin
from stairs.core.app.signals import SignalsMixin
from stairs.core.session import project_session


MODULES_FOR_IMPORT = [
    'app_config',
    'pipelines',
    'producers',
    'consumers',
]


class App(ComponentsMixin, SignalsMixin):

    """
    Used for collect all components and manage everything inside project.
    """

    # app name, extracted from app_package
    app_name = None

    # Aggregator class with all app components
    components = None

    # Project, with full configuration.
    project = None

    def __init__(self, app_name, project=None):

        self.project = project
        if self.project is None:
            self.project = project_session.get_project()

        self.config = AttrDict()

        self.dbs = self.project.dbs
        self.components = components.AppComponents()

        self.app_name = app_name

        self.add_to_project()

        self.send_signal_on_app_created()

    def __dir__(self):
        """
        Me want to make public only components which used in Queue Manager or
        in cli.

        TODO: make better implementation of public components for app object.
        """
        return list(self.components.producers.keys()) + \
               list(self.components.workers.keys())

    def add_to_project(self):
        self.project.add_app(self)


def try_to_import(app_path):
    app_package = importlib.import_module(app_path)

    for module in MODULES_FOR_IMPORT:
        module_path = "%s.%s" % (app_path, module)

        spam_spec = importlib.util.find_spec(module_path)
        found = spam_spec is not None

        if not found:
            continue

        importlib.import_module(module_path)





import os

from flask import Flask, url_for
from flask_mako import MakoTemplates
from flask_mako import render_template

from stairs.services.admin import stairs_config, flask_config

from stairs import get_project

basedir = "/Users/oleg/programming/stairs/stairs/services/admin"

app = Flask(import_name='stairs.services.admin.server',
            root_path=flask_config.basedir,#basedir,
            static_folder='templates/static')

app.config.from_object('stairs.services.admin.flask_config')

mako = MakoTemplates(app)

## Route to any template


@app.route('/')
def index():
    apps = get_project().apps
    return render_template('main.mako',
                           apps=apps)


@app.route('/pipeline/<app_name>/<worker_key>')
def pipeline(app_name, worker_key):
    apps = get_project().apps

    app = [app for app in apps if app.app_name == app_name]
    if len(app) != 1:
        raise RuntimeError("APP not found")

    app = app[0]

    worker = app.components.workers.get(worker_key, None)
    if worker is None:
        raise RuntimeError("Worker not found")

    return render_template('pipeline.mako',
                           apps=apps,
                           app=app,
                           worker=worker)


@app.route('/generator/<app_name>/<generator_key>')
def generator(app_name, generator_key):
    apps = get_project().apps

    app = [app for app in apps if app.app_name == app_name]
    if len(app) != 1:
        raise RuntimeError("APP not found")

    app = app[0]

    producer = app.components.producers.get(generator_key, None)
    if producer is None:
        raise RuntimeError("Producers not found")

    return render_template('generator.mako',
                           apps=apps,
                           app=app,
                           producer=producer)


def run_admin_server():
    # run on port 5000 by default

    port = int(os.environ.get('PORT', 5000))
    app.run(host='0.0.0.0', port=port)




import inspect
import collections

from stairs.core.utils.attrdict import AttrDict

from stairs.core.flow.step import StairsStep


class FlowMeta(type):

    @classmethod
    def __prepare__(metacls, name, bases, **kwds):
        return collections.OrderedDict()

    def __new__(self, name, bases, classdict):
        classdict['__ordered__'] = []

        for base in bases:
            classdict['__ordered__'].extend(base.__ordered__)

        cls_methods = [key for key in classdict.keys()]
        classdict['__ordered__'].extend(cls_methods)

        return type.__new__(self, name, bases, classdict)


class Flow(metaclass=FlowMeta):

    def __reconnect__(self):
        pass

    def compile(self, pipeline):
        self.pipeline = pipeline

        __flow_steps = dict()

        for flow_step_handler_name in self.__ordered__:
            flow_step_handler = getattr(self, flow_step_handler_name, None)

            if not flow_step_handler:
                continue

            handler_meta = getattr(flow_step_handler, '__meta__', None)
            if not handler_meta:
                continue

            flow_step = handler_meta.get('step', None)
            if not flow_step:
                continue

            next_steps = []

            if flow_step.next_steps:
                for next_step in flow_step.next_steps:
                    if next_step is None:
                        next_steps.append(None)
                    else:
                        next_steps.append(__flow_steps[next_step.__name__])

            step = StairsStep(pipeline,
                              self,
                              flow_step.handler,
                              next_steps,
                              save_result=flow_step.save_result,
                              name=flow_step.name)
            __flow_steps[flow_step.handler.__name__] = step
            setattr(self, flow_step.handler.__name__, step)

        self.__reconnect__()

        self.__steps = __flow_steps

    def start_from(self, step, **data):
        return step(**data)

    @classmethod
    def __name__(cls):
        module_name = inspect.getmodule(cls).__name__
        return "%s:%s" % (cls.__name__, module_name)

    @classmethod
    def key(cls):
        return str(cls.__name__)

    @classmethod
    def name(cls):
        return cls.__name__


def get_fields(flow):
    return flow.__dict__.get('__annotations__', {})



import time
from stairs import get_project

from stairs.services.inspect.utils import SimpleQueueStats, MonitorQueueStats


def get_all(app):
    inspects = []

    for pipeline in app.components.workers.values():
        status = get_pipeline_status(pipeline)
        inspects.append(status)

    return inspects


def get_pipeline_status(pipeline, perf_iters_count=4, sleep_time_in_sec=2):
    elps_time = 0
    jobs_amount = 0

    jobs_list = []

    for i in range(perf_iters_count):
        # let's consider jobs count request super fast
        jobs_list.append(get_jobs_count(pipeline))
        if i != perf_iters_count - 1:
            time.sleep(sleep_time_in_sec)

    for i in range(1, perf_iters_count, 1):
        elps_time += sleep_time_in_sec
        jobs_amount += (jobs_list[i]-jobs_list[i-1])

    return SimpleQueueStats(name=pipeline.get_handler_name(),
                            amount_of_jobs=jobs_list[-1],
                            jobs_processed=jobs_amount,
                            amount_of_time=elps_time)


def get_jobs_count(pipeline):
    return get_project().stepist_app.worker_engine.jobs_count(pipeline.step)


def get_from_monitor(app, monitoring_for_sec=10):
    worker_engine = get_project().stepist_app.worker_engine

    step_keys = []
    for pipeline in app.components.workers.values():
        step_keys.append(pipeline.step.step_key())

    s_push, s_pop = worker_engine.monitor_steps(step_keys,
                                                monitoring_for_sec=monitoring_for_sec)

    inspects = []

    for pipeline in app.components.workers.values():
        step_key = pipeline.step.step_key()

        new_jobs = s_push.get(step_key, 0)
        jobs_processed = s_pop.get(step_key, 0)

        s = MonitorQueueStats(name=pipeline.get_handler_name(),
                              amount_of_jobs=get_jobs_count(pipeline),
                              jobs_processed=jobs_processed,
                              amount_of_time=monitoring_for_sec,
                              new_jobs=new_jobs)
        inspects.append(s)

    return inspects





class KeyToKey:

    def __init__(self, transformation_map: dict):
        self.transformation_map = transformation_map

    def __call__(self, data_to_transform):
        output_data = dict()

        for key, value in data_to_transform.items():
            new_key = self.transform_and_assign(key)
            if new_key is not None:
                output_data[new_key] = value

        return output_data

    def transform_and_assign(self, key):

        if not self.transformation_map:
            return key

        if key in self.transformation_map:
            new_key = self.transformation_map[key]
            return new_key


class KeysToDict:
    def __init__(self, dict_key, transformation_map):
        self.dict_key = dict_key
        self.transformation_map = transformation_map
        self.key_to_key = KeyToKey(transformation_map)

    def __call__(self, data_to_transform):
        if self.transformation_map:
            return {self.dict_key: self.key_to_key(data_to_transform)}
        else:
            return {self.dict_key: data_to_transform}


class AllKeys:
    def __init__(self, transformation_map):
        self.transformation_map = transformation_map
        self.key_to_key = KeyToKey(transformation_map)

    def __call__(self, data_to_transform):
        if self.transformation_map:
            return self.key_to_key(data_to_transform)
        else:
            return data_to_transform


import ujson
import random
import time

from stepist.flow.steps.step import StepData


BITMAP_EXPIRE_TIME = 60  # 1 minute


class JobsManager(object):

    def __init__(self, adapter, max_chunks_online, max_jobs_per_input=1000000):
        self.adapter = adapter
        self.max_chunks_online = max_chunks_online

        self.max_jobs_per_input = max_jobs_per_input

    def init_producer_workers(self, cursor_chunks_list):
        """
        Add jobs with chunks id, which should be processed.
        """
        redis_db = self.adapter.app.dbs.redis_db

        count = len(cursor_chunks_list)

        redis_db.setbit(self.worker_checker_key(), count, 0)

    def chunks_producer_worker(self, cursor_chunks_list):
        """
        Start's reading process.
        """

        redis_db = self.adapter.app.dbs.redis_db
        amount_of_chunks = len(cursor_chunks_list)

        while True:
            # getting random chunks
            chunk_index = random.randint(0, amount_of_chunks - 1)

            pipe = redis_db.pipeline()
            # check if we already processed this chunk
            pipe.getbit(self.worker_checker_key(), chunk_index)
            pipe.exists(self.worker_checker_key())

            bit, exist = pipe.execute()

            if int(bit) == 1:
                # checking if we already processed all chunks
                chunks_done = redis_db.bitcount(self.worker_checker_key())
                if chunks_done >= amount_of_chunks:
                    # if we processed everything. Let's remove bits map.
                    redis_db.delete(self.worker_checker_key())
                    print("generator done")
                    exit()
                continue

            if int(exist) == 0:
                print("looks like generator done")
                exit()

            queue_key = self.get_queue_key()

            # trying to get client's list names and set redis name.
            pipe = redis_db.pipeline()
            pipe.client_list()
            pipe.client_setname("chunk_%s" % chunk_index)
            pipe.llen(queue_key)
            clients_list, _, queue_len = pipe.execute()

            # if we already have client which working on current chunk
            # continue workflow
            if "chunk_%s" % chunk_index in [c['name'] for c in clients_list]:
                continue

            if queue_len > self.max_jobs_per_input:
                print("to much jobs, sleeping .. ")
                time.sleep(20)
                continue

            iter = cursor_chunks_list[chunk_index]

            pipe = redis_db.pipeline()

            for row in iter:
                step_data = StepData(flow_data=row)
                data = dict(data=step_data.get_dict())
                pipe.lpush(queue_key, ujson.dumps(data))

            pipe.setbit(self.worker_checker_key(), chunk_index, 1)
            pipe.execute()

    def worker_checker_key(self):
        return "worker_generator:%s" % self.get_queue_key()

    def get_queue_key(self):
        return "stepist::job::%s" % self.adapter.app_input.step.step_key()

    def chunk_worker_key(self,):
        return "%s:chunk_worker" % self.adapter.get_redis_key(),

    def chunk_key(self, chunk_id):
        return "%s:chunk:%s" % (self.adapter.get_redis_key(),
                                chunk_id)



import inspect

from stairs.core.utils import names
from stairs.core.session import config_session


def config():
    def _wrap_config(handler):
        handler_name = names.func_module_name(handler)
        config_session.register_config(handler_name, handler)

        args, _, _, values = inspect.getargvalues(handler)

        args_result = []

        def _wrap_handler(**kwargs):
            for dep_config_handler in args:
                args_result.append(dep_config_handler(**kwargs))

            return handler(*args_result, **kwargs)

        return _wrap_handler

    return _wrap_config



from stairs.core.app import components


class Producer(components.AppProducer):

    def __init__(self, app, adapter):
        self.adapter = adapter

        components.AppProducer.__init__(self, app)

    def __call__(self):
        self.adapter.init_session()

    def process(self):
        self.adapter.process()

    def key(self):
        return self.adapter.handler.__name__




import click
import os

from distutils.dir_util import copy_tree


@click.group()
def project_cli():
    pass


@project_cli.command('project:new')
@click.argument('name')
def make(name):
    """
    Will make simple project, with basic structure, predefined config, and
    sample functions.
    """

    default_project_path = os.path.join(_get_stairs_source_path(),
                                        'default/default_project')
    default_app_path = os.path.join(_get_stairs_source_path(),
                                    'default/default_app')

    path_to_user_project = './%s' % name
    copy_tree(default_project_path, path_to_user_project)
    copy_tree(default_app_path, os.path.join(path_to_user_project, name))

    with open(os.path.join(path_to_user_project, 'config.py'), 'r') as config_f:
        config_body = config_f.read()

    config_body = config_body.format(app_name=name)

    with open(os.path.join(path_to_user_project, 'config.py'), 'w') as config_f:
        config_f.write(config_body)


# utils

def _get_stairs_source_path():
    default_project_path = __file__
    return os.path.dirname(os.path.dirname(os.path.dirname(default_project_path)))



from .app_config import app
from . import pipelines


@app.producer(pipelines.main)
def generate_data():
    for i in range(20):
        yield i


@app.worker_producer(pipelines.main, None)
def generate_data_worker():
    return [map(lambda x: 1, [1, 2, 3])]



import copy


class LikeFunctional:
    copy_on = []

    def __getattribute__(self, name):
        if name != 'copy_on' and name in self.copy_on:
            new_object = copy.copy(self)
        else:
            new_object = self

        return object.__getattribute__(new_object, name)






class AttrDict(dict):

    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError('%s not found' % name)

    def __setattr__(self, name, value):
        self[name] = value

    @property
    def __members__(self):
        return self.keys()



from stairs.core.producer.adapter import BaseProducerAdapter
from .jobs_manager import JobsManager


class IterWorkerAdapter(BaseProducerAdapter):

    app_input = None
    app_output = None

    handler = None

    jobs_manager = None

    def __init__(self, app, handler, app_input, jobs_manager=None):
        self.app = app
        self.handler = handler
        self.app_input = app_input

        if jobs_manager is None:
            jobs_manager = JobsManager(self, 3)

        self.jobs_manager = jobs_manager

    def init_session(self):
        """
        Init producer session
        """
        cursor_chunks_list = list(self.handler())
        self.jobs_manager.init_producer_workers(cursor_chunks_list)

    def process(self):
        cursor_chunks_list = list(self.handler())
        self.jobs_manager.chunks_producer_worker(cursor_chunks_list)



from stairs.core.utils import AttrDict


class Config(AttrDict):

    def load_python(self, filename):
        """Load a config from a python file."""
        if isinstance(filename, str):
            d = dict(self.__dict__)

            with open(filename) as f:
                code = compile(f.read(), filename, 'exec')
                exec(code, d)
        else:
            d = vars(filename)

        self.update(d, skip_reserved=True)

    def load(self, filename):
        """
        guess load type based on filename
        """
        # only python config available right now
        self.load_python(filename)


def init_project_config(filename):
    from stairs import config
    config.load(filename)





import redis

from stairs.core.project.config import ProjectConfig


class DBs:

    def __init__(self, config: ProjectConfig):
        self.config = config

        self.redis_db = init_redis(**config.redis_kwargs)
        self.redis_stats = init_redis(**config.redis_stats_kwargs)

    def reset(self):
        return self.__class__(self.config)


def init_redis(**redis_kwargs):
    return redis.Redis(**redis_kwargs)




from . import global_stairs


def storage():
    if not hasattr(global_stairs, 'project'):
        global_stairs.project = dict(
            project=None,
        )

    return global_stairs.project


def set_project(project):
    storage()['project'] = project


def get_project():
    return storage()['project']





class BaseProducerAdapter(object):

    app_input = None
    app_output = None

    app = None
    handler = None

    def init_session(self):
        raise NotImplemented()

    def process(self):
        raise NotImplemented()

    def continue_process(self):
        raise NotImplemented()

    def get_redis_key(self):
        return "producer:%s:%s" % (self.app.app_name,
                                   self.handler.__name__)




from . import local


def storage():
    if not hasattr(local, 'step_storage'):
        local.step_storage = dict(
            steps={},
            last_registered_step=None,
        )
    return local.step_storage


def set_step(step):
    storage()['steps'][step.key()] = step


def get_step(step_key):
    return storage()['steps'].get(step_key, None)


def get_last_registered_step():
    return storage().get("last_registered_step", None)


def set_last_registered_step(step):
    storage()['last_registered_step'] = step



from stairs.core.producer.adapter import iter_adapter, iter_worker_adapter
from stairs.core.producer import Producer
from stairs.core.app.components import AppWorker
from stairs.core.output.output_model import Output
from stairs.core.output.standalone import StandAloneConsumer
from stairs.core.output.consumer_iter import ConsumerIter
from stairs.core.worker.worker import Worker


class ComponentsMixin:
    components = None

    def producer(self, *app_input):
        def _handler_wrap(handler):
            if isinstance(app_input, AppWorker):
                app_inputs = [app_input]
            else:
                app_inputs = app_input

            adapter = iter_adapter.IterAdapter(self, handler, app_inputs)
            return Producer(self, adapter=adapter)

        return _handler_wrap

    def worker_producer(self, app_input, jobs_manager=None):
        def _handler_wrap(handler):
            adapter = iter_worker_adapter.IterWorkerAdapter(app=self,
                                                            handler=handler,
                                                            app_input=app_input,
                                                            jobs_manager=jobs_manager)
            return Producer(self, adapter=adapter)

        return _handler_wrap

    def pipeline(self, config=None):
        def _deco_init(func):
            return Worker(self, func, config)

        return _deco_init

    def consumer(self):
        def _handler_wrap(func):
            return Output(app=self, handler=func)

        return _handler_wrap

    def standalone_consumer(self):
        def _handler_wrap(func):
            return StandAloneConsumer(app=self, handler=func)

        return _handler_wrap

    def consumer_iter(self):
        def _handler_wrap(func):
            return ConsumerIter(app=self, handler=func)

        return _handler_wrap

    def get_pipeline(self, name):
        return self.components.workers.get(name)

    def get_producer(self, name):
        return self.components.producers.get(name)

    def get_consumer(self, name):
        return self.components.consumers.get(name)



# import os
import click

from stairs.services.management.project import project_cli
from stairs.services.management.workers import workers_cli
from stairs.services.management.producer import producer_cli
from stairs.services.management.general import general_cli
from stairs.services.management.consumer import consumer_cli
from stairs.services.management.shell import shell_cli
from stairs.services.management.inspect import inspect_cli


def init_cli():
    cli = click.CommandCollection(sources=[producer_cli,
                                           workers_cli,
                                           general_cli,
                                           project_cli,
                                           consumer_cli,
                                           shell_cli,
                                           inspect_cli])
    cli()


if __name__ == "__main__":
    init_cli()



from stairs.core.session import project_session

from stairs.core.output.output_model import Output
from stairs.core.app import components


class ConsumerIter(Output):

    def __init__(self, app, handler, as_daemon=False):
        self.handler = handler
        self.app = app

        self.step = self.app \
            .project \
            .stepist_app \
            .step(None, as_worker=True)(self.run_job)

        components.AppOutput.__init__(self, self.app)

    def __call__(self, *args, **kwargs):
        for data in self.jobs_iterator():
            yield self.handler(**data)

    def jobs_iterator(self):
        """
        The way how it works a bit tricky.
        We manually grab data from worker engine, but we can't just return
        it because stepist session is not initialized.
        For this reason we manually executing step and then just return
        result of this step (result of run_job)
        """
        project = project_session.get_project()

        while True:
            job = project.stepist_app.worker_engine.receive_job(self.step)
            data = self.step.receive_job(**job)
            # it's last step in step and it will be dict with "run_job" key
            yield data.get("run_job")

    def run_job(self, **job_data):
        return job_data



import click
from stairs.core import app

@click.group()
def general_cli():
    pass


@general_cli.command("run")
@click.argument('name')
def init_session(name):
    app_name, command_name = name.split('.')
    app.get_app_by_name(app_name).get_command(command_name)()


@general_cli.command("admin")
def init_admin():
    from stairs.services.admin import server
    server.run_admin_server()



from . import local

from contextlib import contextmanager


def storage():
    if not hasattr(local, 'pipeline_storage'):
        local.pipeline_storage = dict(
            pipeline=None,
        )
    return local.pipeline_storage


@contextmanager
def change_pipeline_ctx(pipeline):
    prev_pipeline = storage().get("pipeline")

    try:
        set_pipeline(pipeline)
        yield
    finally:
        set_pipeline(prev_pipeline)


def set_pipeline(pipeline):
    storage()['pipeline'] = pipeline


def get_pipeline():
    return storage().get('pipeline', None)





class ComponentContext:
    """
    Provide context for component, which data should be transferred and who
    will be responsible for this data in the future.
    """

    def __init__(self, p_component, component_keys_transformation):
        """
        :param p_component:
        :param transformation: key:value dict, where:
            key - data value name which we want to transfer
            value - data value name which should be used in new component
        """
        self.to_p_component = p_component
        self.transformation = component_keys_transformation

    def assign_labels(self, data):
        output_data = dict()

        for key, value in data.items():
            key = "%s->%s" % (key, self.to_p_component.id)
            output_data[key] = value

        return output_data


def is_context_key(key):
    return '->' in key


def belongs_to_component(component, key):
    if '->' in key:
        return component.id in key

    return False


def unassign_data(component, key):
    return key.split('->')[0]



from . import local
from contextlib import contextmanager


def storage():
    if not hasattr(local, 'worker_storage'):
        local.worker_storage = dict()

    return local.worker_storage


def set_worker(worker):
    storage()['current_worker'] = worker


def get_worker():
    return storage().get('current_worker', None)


@contextmanager
def change_worker_ctx(worker):
    prev_worker = get_worker()

    try:
        set_worker(worker)
        yield
    finally:
        set_worker(prev_worker)



import click

from stairs import get_project
from stairs.services.inspect.pipelines import get_from_monitor, get_all


@click.group()
def inspect_cli():
    pass


@inspect_cli.command("inspect:status")
@click.argument("app_name", default=None)
def init_session(app_name):
    project = get_project()

    if app_name:
        app = project.get_app_by_name(app_name)
        apps = [app]
    else:
        apps = get_project().apps

    for app in apps:
        inspection = get_all(app)
        for status in inspection:
            status.print_info()
            print("=========")


@inspect_cli.command("inspect:monitor")
@click.argument("app_name", default=None)
def init_session(app_name):
    project = get_project()

    if app_name:
        app = project.get_app_by_name(app_name)
        apps = [app]
    else:
        apps = get_project().apps

    for app in apps:
        inspection = get_from_monitor(app)
        for status in inspection:
            status.print_info()
            print("=========")




from stepist.flow.utils import StopFlowFlag

from .core.flow.flow import Flow
from .core.flow.step import step
from .core.utils.execeptions import StopPipelineFlag
from .core.worker.data_pipeline import concatenate

from .core.app import App

from .core.session.project_session import get_project



import queue


class GraphItem:

    def __init__(self, pipeline_component, next=None):
        self.p_component = pipeline_component

        self.id = pipeline_component.id
        self.next = next if next else []

    def __eq__(self, other):
        return self.p_component.id == \
               other.p_component.id

    def __str__(self):
        return str(self.id)

    def deepcopy(self, ):
        item = self.__class__(self.p_component, next=[])

        for n in self.next:
            item.next.append(n.deepcopy())
        return item

    def add_next(self, base_item, with_deep_copy=False):
        # it's important to create new Item instance,
        # because otherwise different graph items will have same 'next' values
        if with_deep_copy:
            self.next.append(base_item.deepcopy())
        else:
            self.next.append(base_item)

    def add_next_graph(self):
        pass


class PipelineGraph:
    """
    Graph which made specialy for Pipeline objects.
    This graph is actually a tree:
    (graph which don't have cycles and amount of edges = amount of items - 1).

    Some terms:
        Leave - graph item which don't have "next" items
        Level - the distance between root and some item
        Max Level - the depth of the Graph.


    """

    def __init__(self, root=None):
        self.root = root

    def __str__(self):
        result_items = []
        for item in dfs(self.root):
            result_items.append(str(item))

        return str(result_items)

    def deepcopy(self):
        return PipelineGraph(root=self.root.deepcopy())

    def get_root(self):
        """
        Return graph root. (The first added item to the graph)
        """
        return self.root

    def is_line_tree(self):
        """
        "Line tree" - is a tree which have only one leave.

        :return: True if graph (tree) has one leave only
        """

        return len(get_leaves(self)) == 1

    def extend_leaves(self, new_item):
        """
        Add new graph item to all defined level items.

        Looks like:

        If we have :

                (2) -> (4)
        (1) ->
                (3) -> (5)

        And we want to add (6) on all leaves

                (2) -> (4) -> (6.1)
        (1) ->
                (3) -> (5) -> (6.2)

        """

        if self.root is None:
            self.root = new_item.deepcopy()
            return

        for item in get_leaves(self):
            if not item.next:
                item.add_next(new_item)

    def add_pipeline_component(self, pipeline_component):
        graph_item = GraphItem(pipeline_component)

        leaves_affected = get_leaves(self)
        self.extend_leaves(graph_item)

        return leaves_affected

    def add_graphs_on_leaves(self, *new_graphs):
        """
        TODO: documentation
        """

        current_leaves = get_leaves(self)

        if not current_leaves:
            raise RuntimeError("You can't add multiple graphs to empty graph")

        for item in current_leaves:
            for new_graph in new_graphs:
                self.add_graph_root_to_item(item, new_graph.root)

        return current_leaves

    def add_graph_on_leaves(self, graph):
        # syntax sugar function
        # mockup under add_graphS_on_leaves

        return self.add_graphs_on_leaves(*[graph])

    def add_graph_root_to_item(self, parrent_item, new_graph_root):
        parrent_item.add_next(new_graph_root)

# graph utils:

def get_leaves(graph):
    if graph.root is None:
        return []

    return get_leaves_graph_item(graph.root)


def get_leaves_graph_item(g_item):
    return [item for item in dfs(g_item) if not item.next]


def dfs(current_item, backward=False):
    """
    Go through all graph in recursion way.
    See https://en.wikipedia.org/wiki/Depth-first_search for more info

    :param current_item: Item which we should start from
    :param backward: The way how we should return results

    :yield: graph items in dfs way.
    """
    if not backward:
        yield current_item

    for next_item in current_item.next:
        for item in dfs(next_item, backward=backward):
            yield item

    if backward:
        yield current_item


def dfs_by_multiple(item1, item2, backward=False):
    """
    Same like general dfs. But we iteration only by similar vertex from both
    graphs.
    """

    if item1 != item2:
        return

    if not backward:
        yield item1, item2

    for next_item1 in item1.next:
        for next_item2 in item2.next:
            if next_item1 != next_item2:
                continue

            for item1, item2 in dfs_by_multiple(next_item1, next_item2,
                                                backward=backward):
                yield item1, item2

    if backward:
        yield  item1, item2


def bfs(current_item):
    """
    Go through all graph using queue (FIFO)
    See https://en.wikipedia.org/wiki/Breadth-first_search for more info

    :yield: graph items in bfs way
    """
    q = queue.Queue()
    q.put(current_item)

    while not q.empty():
        g_item = q.get()
        yield g_item

        for next_item in g_item.next:
            q.put(next_item)


def concatenate_sequentially(graph1, graph2):
    """
    Combine two different graphs with same root in "line" tree.
    Both graphs should be "line" tree. (one vertex - one child)

    If we have two following graphs:

    1)

    (1) -> (2) -> (3) -> (4)

    2)

    (1) -> (2) -> (5) - (6)

    We should merge them into:

    (1) -> (2) -> (3) -> (4) -> (5) -> (6)

    And add data transformation function to min parent leave:

    (1) -> (2) -> (3) -> (4) -> (5) -> (6)
            \                   ^
              -  - - - - - - - /
                data from (2)

    :return graph item which need to be extend by transformation function
    """

    if not graph1.is_line_tree() or not graph2.is_line_tree():
        pass

    graph = merge_graphs_by_roots(graph1, graph2)
    leaves = get_leaves(graph)

    # if we have only one leave -> g2 is subset of g1
    if len(leaves) == 1:
        return graph, None, None

    # if both graphs are inline, we should always have 2 leaves after merging
    if len(leaves) != 2:
        raise RuntimeError("Something wrong with graph after concatinate, "
                           "looks like it's not a 'line' tree ")

    for g_item in dfs(graph.root):
        if len(g_item.next) == 2:
            left_item = g_item.next[0]
            right_item = g_item.next[1]

            left_leave = get_leaves_graph_item(left_item)[0]

            # add connection to next
            left_leave.add_next(right_item)

            # remove connection to right item
            g_item.next = [left_item]

            return graph, g_item, right_item


def merge_graphs_by_roots(graph1, graph2):
    """
    Populate graph1 by graph2 items. Consider only none-equal elements.
    """

    if graph1.root != graph2.root:
        raise RuntimeError("Graphs has different roots")

    for g1_item, g2_item in dfs_by_multiple(graph1.root, graph2.root):
        different_items_from_g2 = []
        # Extract all items from g2 which different from g1 on g1_item level.
        for next_item_g2 in g2_item.next:
            g1_has_item_from_g2 = False

            for next_item_g1 in g1_item.next:
                if next_item_g1 == next_item_g2:
                    g1_has_item_from_g2 = True

            if not g1_has_item_from_g2:
                different_items_from_g2.append(next_item_g2.deepcopy())

        g1_item.next.extend(different_items_from_g2)

    return graph1



import importlib

from stairs.core.utils import AttrDict
from stairs.core.session import step_session

MODULES_FOR_IMPORT = [
    'app_config',
    'pipelines',
    'producers',
    'consumers',
]


class AppBaseComponent(object):
    """
    Base app component - class for specifying basic components methods.

    Each app component, should have stepist initialization, and object ref to
    stepist.step
    """
    app = None
    component_type = None

    def get_stepist_step(self):
        raise NotImplementedError()

    def key(self):
        raise NotImplementedError()


class AppComponents:
    """
    Aggregator of all app components.

    Also playing a role of components auto initialization (by importing modules
    inside the app)
    """

    class Components(AttrDict):

        def add_component(self, component):
            self[component.key()] = component

    def __init__(self):
        self.producers = self.Components()
        self.workers = self.Components()
        self.consumers = self.Components()
        self.flows = self.Components()
        self.steps = self.Components()

    def try_import(self, base_path):
        for module in MODULES_FOR_IMPORT:
            try:
                importlib.import_module("%s.%s" % (base_path, module))
            except ImportError as e:
                print(e)


class AppWorker(AppBaseComponent):
    def __init__(self, app):
        self.app = app
        app.components.workers.add_component(self)


class AppProducer(AppBaseComponent):
    def __init__(self, app):
        self.app = app
        app.components.producers.add_component(self)


class AppStep(AppBaseComponent):
    def __init__(self, app):
        self.app = app
        app.components.steps.add_component(self)
        step_session.set_last_registered_step(self)


class AppOutput(AppBaseComponent):
    def __init__(self, app):
        self.app = app
        app.components.consumers.add_component(self)


class AppFlow(AppBaseComponent):
    def __init__(self, app):
        self.app = app
        app.components.flows.add_component(self)



import inspect

from stepist.flow.steps.next_step import call_next_step

from stairs.core.app import components
from stairs.core.worker import data_pipeline


class WorkerInfo:

    def __init__(self, base_worker, config=None):
        self.base_worker = base_worker
        self.app = base_worker.app
        self.config = config

    def key(self):
        return self.base_worker.key()


class Worker(components.AppWorker):
    def __init__(self, app, pipeline_builder, worker_config):
        self.pipeline_builder = pipeline_builder
        self.app = app

        self.config = worker_config
        self.pipeline = self.make_pipeline(worker_config)
        self.compile()

        self.step = self.app\
            .project\
            .stepist_app\
            .step(None,
                  as_worker=True,
                  unique_id=self.get_handler_name())(self)

        components.AppWorker.__init__(self, self.app)

    def __call__(self, **kwargs):
        if not self.pipeline.is_compiled():
            raise RuntimeError("Worker pipeline no compiled, "
                               "run worker.compile()")

        result = call_next_step(kwargs, self.pipeline.get_stepist_root())

        if result:
            return list(result.values())[0]

    def compile(self):
        self.pipeline.compile()

    def make_pipeline(self, config=None):
        if config is None:
            # TODO: probably we need to update self.config by passed config
            config = self.config

        worker_info = WorkerInfo(self, config=config)

        initial_pipeline = data_pipeline.DataPipeline.make_empty(self.app,
                                                                 worker_info)
        initial_frames = self.initial_data_frames(initial_pipeline)

        worker_data_frame = self.pipeline_builder(worker_info, **initial_frames)
        ensure_worker_output_is_valid(worker_data_frame)

        return worker_data_frame.data_pipeline

    def key(self):
        return self.get_handler_name()

    def get_worker_input_values(self):
        info = inspect.getfullargspec(self.pipeline_builder)
        args = info[0]
        # exclude worker var
        args.pop(0)
        return args

    def initial_data_frames(self, pipeline):
        worker_input_keys = self.get_worker_input_values()

        initial_frames = dict()
        for key in worker_input_keys:
            data_point = data_pipeline.DataPoint(pipeline, {key: key})
            initial_frames[key] = data_point

        return initial_frames

    def get_stepist_step(self):
        return self.step

    def flush_queue(self):
        # TODO: flush all jobs in stepist
        pass

    def get_handler_name(self):
        return self.pipeline_builder.__name__

    def __name__(self):
        module_name = inspect.getmodule(self.pipeline_builder).__name__
        return "%s:%s:%s" % (self.app.app_name,
                             module_name, self.get_handler_name())


def ensure_worker_output_is_valid(w_output):

    # if len(list(w_output)) != 1:
    #     raise RuntimeError("Worker should return one DataFrame")

    if w_output is None:
        raise RuntimeError("Worker output can't be None")



import inspect

from typing import List

from stepist.app import App as StepistApp
from stepist.flow.steps.step import Step as StepistStep

from stairs.core import app as stairs_app
from stairs.core.app import App as StairsApp
from stairs.core.app.components import AppBaseComponent

from stairs.core.session import project_session

from stairs.core.project import dbs, config as stairs_config
from stairs.core.project import utils


class StairsProject:

    def __init__(self,  stepist_app=None, config_file=None, **config):
        self.apps = []

        if config_file:
            config_file = stairs_config.ProjectConfig.load_from_file(config_file)
            config.update(config_file)

        self.config = stairs_config.ProjectConfig.init_default()
        self._update_config(**config)

        if stepist_app is None:
            self.stepist_app = StepistApp(**self.config)
        else:
            self.stepist_app = stepist_app

        self.dbs = dbs.DBs(self.config)
        project_session.set_project(self)

    @classmethod
    def init_from_file(cls, config_path: str, **init_kwargs):
        config = stairs_config.ProjectConfig.load_from_file(config_path)
        return cls(**init_kwargs, **config)

    def _update_config(self, **config) -> None:
        """
        Updating current config with a new one.

        Should be executed before Stepist App initialized
        """
        # update with custom config
        self.config = stairs_config.ProjectConfig(**{**self.config, **config})
        self.init_apps()

    def run_pipelines(self, die_when_empty=False, die_on_error=True) -> None:
        """
        Iterating over apps and running (listen for jobs) pipelines
        """
        self.stepist_app.run(self.pipelines_to_run(),
                             die_on_error=die_on_error,
                             die_when_empty=die_when_empty)

    def add_app(self, app: StairsApp) -> None:
        """
        Register new app in the project
        """
        self.apps.append(app)

    def init_apps(self) -> None:
        """
        Iterates over all apps in config and try to import them with all
        components inside
        """
        self.apps = []
        if self.config.get('apps', None):
            for app in self.config.apps:
                stairs_app.try_to_import(app)

    def get_app(self, name) -> StairsApp:
        """
        Just shortcust for get_app_by_name
        """
        return self.get_app_by_name(name)

    def get_app_by_name(self, name: str):
        for app in self.apps:
            if app.app_name == name:
                return app

        raise RuntimeError("App '%s' not found" % name)

    def get_app_by_obj(self, obj) -> StairsApp:
        """
        Trying to guess app by some pyobject
        """
        if isinstance(obj, AppBaseComponent):
            return obj.app

        if 'app' in obj.__dict__ and isinstance(obj.app, StepistApp):
            return obj.app

        module_path = inspect.getmodule(obj)
        for app in self.apps:
            if app.app_name in module_path.__name__:
                return app

        raise RuntimeError("App not found for function '%s'" % obj.__name__)

    def pipelines_to_run(self) -> List[StepistStep]:
        components_to_run = []

        for step in self.stepist_app.get_workers_steps():
            related_to_pipeline = False

            for pipeline_step in utils.PIPELINES_STEPS_TO_RUN:
                if isinstance(step.handler, pipeline_step):
                    related_to_pipeline = True

            if related_to_pipeline:
                components_to_run.append(step)

        return components_to_run




import click

from IPython.terminal.embed import InteractiveShellEmbed


@click.group()
def shell_cli():
    pass


@shell_cli.command("shell")
def init_session():
    shell = InteractiveShellEmbed()
    shell()



import copy
import uuid
from stepist.flow.steps.next_step import call_next_step as stepist_next_step

from stairs.core.worker.pipeline_objects import (PipelineFlow,
                                                 PipelineFlowProducer,
                                                 PipelineOutput,
                                                 PipelineFunction,
                                                 PipelineInVainComponent)

from stairs.core.worker import pipeline_graph
from stairs.core.worker.pipeline_graph import concatenate_sequentially

from stairs.core.worker.pipeline_objects import \
    transformation as transformations_types


class DataPipeline:
    """

    """

    def __init__(self, app, worker_info, graph=None, initial_component=None):
        self.app = app
        self.worker_info = worker_info

        # if graph not define, trying to make "empty" graph
        if graph is None:
            graph = pipeline_graph.PipelineGraph()

            if initial_component is None:
                initial_component = self.get_initial_component(worker_info)

            graph.add_pipeline_component(initial_component)

        self.graph = graph
        self.stepist_steps = dict()

    def __call__(self, *args, **kwargs):
        if not self.stepist_steps:
            raise RuntimeError("Pipeline not compiled")

        stepist_next_step(self.get_stepist_root(), **kwargs)

    def deepcopy(self):
        cls = self.__class__

        return cls(self.app,
                   worker_info=self.worker_info,
                   graph=self.graph.deepcopy())

    def keys(self):
        if not self.graph.is_line_tree():
            raise RuntimeError("Can't get keys from graph with multiple leaves")

        leave = pipeline_graph.get_leaves(self.graph)[0]
        return leave.p_component.get_keys()

    def add_pipeline_component(self, p_component, transformation:dict):
        graph_items_affected = self.graph.add_pipeline_component(p_component)

        for g_item in graph_items_affected:
            g_item.p_component.add_context(p_component,
                                           transformations_types.KeyToKey(transformation))

        return self

    def add_pipeline_graph(self, graph, transformation):
        graph_items_affected = self.graph.add_graph_on_leaves(graph)

        root_p_component = graph.get_root().p_component

        for g_item in graph_items_affected:
            g_item.p_component.add_context(
                root_p_component,
                transformations_types.KeyToKey(transformation)
            )

    def get_last_item(self):
        return pipeline_graph.get_leaves(self.graph)[0]

    def compile(self):
        if self.is_compiled():
            raise RuntimeError("Pipeline already compiled")

        self.stepist_steps = compile_pipeline(self)

    def is_compiled(self):
        return bool(self.stepist_steps)

    def get_stepist_id(self, component_id):
        return "%s:%s" % (self.worker_info.key(), component_id)

    # utils:
    def get_stepist_root(self):
        root_component = self.graph.get_root().p_component
        stepist_id = self.get_stepist_id(root_component.id)

        return self.stepist_steps[stepist_id]

    @classmethod
    def make_empty(cls, app, worker_info):
        return cls(
            app,
            worker_info,
            initial_component=cls.get_initial_component(worker_info)
        )

    @classmethod
    def get_initial_component(cls, worker_info):
        return PipelineInVainComponent(
            pipeline=None,
            name="%s:root" % worker_info.key()
        )


class DataFrame:

    def __init__(self, data_pipeline: DataPipeline, transformation: dict=None,
                 possible_keys=None):

        self.data_pipeline = data_pipeline
        self.transformation = transformation or {}
        self.possible_keys = possible_keys or []

    def make(self, *keys, **transformation):
        # populate transformation by keys
        for key in keys:
            transformation[key] = key

        for key, value in transformation.items():
            if isinstance(value, DataPoint):
                transformation[key] = value.get_key()

        transformation = {v: k for k, v in transformation.items()}
        transformation = self.update_by_current_transformation(transformation)

        return self.__class__(self.data_pipeline, transformation)

    def get(self, key):
        return DataPoint(self.data_pipeline, {key: key})

    def update_by_current_transformation(self, new_transformation) -> dict:
        for key, value in self.transformation.items():
            if value in new_transformation:
                new_transformation[key] = new_transformation[value]
                del new_transformation[value]
            else:
                new_transformation[key] = value

        return new_transformation

    def has_one_transformation(self):
        return len(self.transformation) == 1

    def subscribe_flow(self, flow, as_worker=False, update_pipe_data=True,
                       name=None):
        data_pipeline = self.data_pipeline.deepcopy()

        name = name or "%s:%s" % (self.data_pipeline.worker_info.key(),
                                  flow.name())
        config = data_pipeline.worker_info.config

        p_component = PipelineFlow(self.data_pipeline,
                                   flow,
                                   as_worker=as_worker,
                                   name=name,
                                   config=config,
                                   update_pipe_data=update_pipe_data)

        data_pipeline.add_pipeline_component(
            p_component,
            transformation=self.transformation
        )

        return DataFrame(data_pipeline)

    def subscribe_flow_as_producer(self, flow, as_worker=False, name=None,
                                   update_pipe_data=True):
        data_pipeline = self.data_pipeline.deepcopy()

        name = name or "%s:%s" % (self.data_pipeline.worker_info.key(),
                                  flow.name())
        config = data_pipeline.worker_info.config

        p_component = PipelineFlowProducer(self.data_pipeline,
                                           flow,
                                           as_worker=as_worker,
                                           name=name,
                                           config=config,
                                           update_pipe_data=update_pipe_data)

        data_pipeline.add_pipeline_component(
            p_component,
            transformation=self.transformation
        )

        return DataFrame(data_pipeline)

    def subscribe_pipeline(self, app_worker, config=None, update_pipe_data=True):
        data_pipeline = self.data_pipeline.deepcopy()

        pipeline = app_worker.make_pipeline(config=config)
        pipeline.graph.get_root().p_component.as_worker = True

        data_pipeline.add_pipeline_graph(pipeline.graph,
                                         self.transformation)

        return DataFrame(data_pipeline)

    def subscribe_consumer(self, output, name=None, as_worker=False):
        data_pipeline = self.data_pipeline.deepcopy()

        name = name or "%s:%s" % (self.data_pipeline.worker_info.key(),
                                  output.name())
        p_component = PipelineOutput(
            self.data_pipeline,
            name=name,
            component=output,
            config=data_pipeline.worker_info.config,
            as_worker=as_worker,
            update_pipe_data=False
        )

        data_pipeline.add_pipeline_component(
            p_component,
            transformation=self.transformation
        )

        return DataFrame(data_pipeline)

    def subscribe_func(self, func, as_worker=False, name=None,
                       update_pipe_data=True):
        data_pipeline = self.data_pipeline.deepcopy()

        name = name or "%s:%s" % (self.data_pipeline.worker_info.key(),
                                  func.__name__)
        config = data_pipeline.worker_info.config

        p_component = PipelineFunction(self.data_pipeline,
                                       func,
                                       as_worker=as_worker,
                                       name=name,
                                       config=config,
                                       update_pipe_data=update_pipe_data)

        data_pipeline.add_pipeline_component(
            p_component,
            transformation=self.transformation
        )

        return DataFrame(data_pipeline)

    def apply_flow(self, flow, as_worker=False,):
        return self.subscribe_flow(flow=flow,
                                   as_worker=as_worker,
                                   update_pipe_data=False)

    def apply_pipeline(self, app_worker, config=None,
                       update_pipe_data=True):

        return self.subscribe_pipeline(app_worker, config=config,
                                       update_pipe_data=update_pipe_data)

    def add_value(self, **kwargs):
        func = lambda **k: kwargs

        # it's safe to use uuid4 as name, because it can't be a worker
        return self.subscribe_func(func,
                                   name=uuid.uuid4(),
                                   as_worker=False)


class DataPoint(DataFrame):
    def get_key(self):
        return list(self.transformation.keys())[0]


def compile_pipeline(pipeline):
    """
    Go through all graph nodes, and retranslate them to stepist graph.
    Should be executed once for one pipeline (pipeline component). Otherwise
    Stepist will raise Error about handler duplication


    :param pipeline: pipeline with graph object.
    :return: dict (map) stepist_unique_id: stepist.step
    """

    stepist_app = pipeline.app.project.stepist_app
    stepist_steps = dict()
    dfs_iteration = pipeline_graph.dfs(pipeline.graph.get_root(), backward=True)

    for i, graph_item in enumerate(dfs_iteration):
        unique_id = pipeline.get_stepist_id(graph_item.p_component.id)
        graph_item.p_component.stepist_id = unique_id

        if len(graph_item.next) == 0:
            step = stepist_app.step(
                None,
                as_worker=graph_item.p_component.as_worker,
                unique_id=unique_id,
                name=unique_id,
            )(graph_item.p_component)

        elif len(graph_item.next) == 1:
            next = graph_item.next[0]
            next_stepist = stepist_steps.get(
                pipeline.get_stepist_id(next.p_component.id)
            )

            step = stepist_app.step(
                next_stepist,
                as_worker=graph_item.p_component.as_worker,
                unique_id=unique_id,
                name=unique_id,
            )(graph_item.p_component)

        else:
            raise RuntimeError("next more than one, implement map support")

        stepist_steps[unique_id] = step

    return stepist_steps


def concatenate(*data_frames, **data_points):
    """

    """
    #ensure_concatenate_allowed(data_points)

    base_pipeline = None
    if data_frames:
        base_pipeline = data_frames[0].data_pipeline
    if data_points:
        base_pipeline = list(data_points.values())[0].data_pipeline

    invain_name = "%s:%s" % ("concatenate", ",".join(data_points.keys()))
    last_p_component = PipelineInVainComponent(
        pipeline=base_pipeline,
        name=invain_name)

    # TODO: check if all data_pipelines_transformations actually transformations

    data_pipeline_items = [dp.data_pipeline for dp in data_points.values()]

    # generate's result transformation object
    for key, data_point in data_points.items():

        p = data_point.data_pipeline
        leaves = pipeline_graph.get_leaves(p.graph)

        assert len(leaves) == 1
        if isinstance(data_point, DataPoint):
            keys_to_transform = {data_point.get_key(): key}
            transformation_func = transformations_types.KeyToKey(keys_to_transform)
            leaves[0].p_component.add_context(last_p_component,
                                              transformation_func)
        else:
            keys_to_transform = data_point.transformation
            transformation_func = transformations_types.KeysToDict(key,
                                                            keys_to_transform)
            leaves[0].p_component.add_context(last_p_component,
                                              transformation_func)

    for data_frame in data_frames:
        p = data_frame.data_pipeline
        leaves = pipeline_graph.get_leaves(p.graph)
        assert len(leaves) == 1
        keys_to_transform = data_frame.transformation
        transformation_func = transformations_types.AllKeys(keys_to_transform)
        leaves[0].p_component.add_context(last_p_component,
                                          transformation_func)

    base_data_pipeline = copy.copy(data_pipeline_items[0])
    base_graph = base_data_pipeline.graph
    base_worker_info = data_pipeline_items[0].worker_info

    for data_pipeline in data_pipeline_items[1:]:

        if data_pipeline.worker_info is not base_worker_info:
            raise RuntimeError("Not possible to use different worker configs, "
                               "for pipeline")

        graph, base_g_item, handler_item = \
            concatenate_sequentially(base_graph, data_pipeline.graph)

        base_graph = graph

    base_graph.add_pipeline_component(last_p_component)

    return DataFrame(base_data_pipeline)


def ensure_subscribe_allowed(transformation):
    for key, data_pipeline in transformation.items():
        graph_leaves = pipeline_graph.get_leaves(data_pipeline.graph)

        if not graph_leaves[0].p_component.get_one_value:
            raise RuntimeError("Found multiple values for "
                               "one transformation key")


def ensure_concatenate_allowed(transformation):

    for key, data_pipeline in transformation.items():
        graph_leaves = pipeline_graph.get_leaves(data_pipeline.graph)

        if len(graph_leaves) > 1:
            raise RuntimeError("Can't concatenate multiple leaves graph")

        # if not graph_leaves[0].pipeline_component.get_one_value:
        #     raise RuntimeError("Found multiple values for "
        #                        "one transformation key")



import click
from stairs.core.session import project_session


@click.group()
def producer_cli():
    pass


@producer_cli.command("producer:init_session")
@click.argument('name')
def init_session(name):
    project = project_session.get_project()
    print("init producer session")
    app_name, producer_name = name.split('.')
    user_app = project.get_app_by_name(app_name)
    user_app.components.producers[producer_name]()


@producer_cli.command("producer:process")
@click.argument("name")
def process(name):
    project = project_session.get_project()
    print("init producer session")
    app_name, producer_name = name.split('.')
    user_app = project.get_app_by_name(app_name)
    print(user_app.components.producers)
    user_app.components.producers[producer_name].process()



from contextlib import contextmanager

from stairs.core.session import local


def flow_storage():
    if not hasattr(local, 'local_flow'):
        local.local_flow = dict()

    return local.local_flow


def get_flow():
    return flow_storage().get("flow", None)


def get_step_result(step_key):
    s = flow_storage()

    return s.get('flow_steps', {}).get(step_key, None)


def get_result_steps():
    s = flow_storage()

    return s.get('result_steps', None)


def flush_step_results():
    s = flow_storage()
    s['flow_steps'] = {}


def set_flow(flow_instance):
    flow_storage()["flow"] = flow_instance


def set_flow_step(flow_step, step_data):
    s = flow_storage()

    if "flow_steps" not in s:
        s["flow_steps"] = dict()

    s['flow_steps'][flow_step.key()] = step_data


def set_result_steps(result_steps):
    s = flow_storage()
    s['result_steps'] = result_steps


@contextmanager
def change_result_steps(result_steps):
    prev_steps = get_result_steps()

    try:
        set_result_steps(result_steps)
        yield
    finally:
        set_result_steps(prev_steps)


@contextmanager
def change_flow_ctx(flow):
    prev_flow = get_flow()

    try:
        set_flow(flow)
        yield
    finally:
        set_flow(prev_flow)



from setuptools import setup, find_packages


setup(
    name="stairs-project",
    version="0.1",
    author="Aleh Shydlouski",
    author_email="oleg.ivye@gmail.com",
    description="Framework for data processing using data pipelines",
    keywords=['data-pipeline',
              'data-science',
              'data-engineering',
              'distributed-computing'],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'tqdm==4.25.0',
        'redis==3.0.1',
        'blinker==1.4',
        'click==7.0',
        'ujson==1.35',
        'flask',
        'stepist'
    ],
    url='https://github.com/electronick1/stairs',
    download_url='https://github.com/electronick1/stairs/archive/0.0.3.tar.gz',
    classifiers=[],
    entry_points={'console_scripts': [
        'stairs-admin = stairs.services.management.commands:init_cli',
    ]},
)





class SignalsMixin:
    signals_on_app_created = []

    def on_app_created(self):
        def _wrap_handler(handler):
            self.signals_on_app_created.append(handler)

        return _wrap_handler

    def send_signal_on_app_created(self):
        for handler in self.signals_on_app_created:
            handler(self)



from stairs.core.flow import Flow, step


class FlowSample(Flow):

    def __call__(self, data):
        r = self.start_from(self.first_step, value=data)

        return dict(output=r.second_step)

    @step(None)
    def second_step(self, value2):
        return dict(value3=value2)

    @step(second_step)
    def first_step(self, value):
        return dict(value2=value)



import types

from stepist.flow.steps.next_step import call_next_step
from stepist.flow.utils import validate_handler_data, StopFlowFlag

from stairs.core.utils.execeptions import StopPipelineFlag

from stairs.core.worker.pipeline_objects import context as pipeline_context
from stairs.core.session import unique_id_session


class PipelineComponent:

    def __init__(self, pipeline, component, name, config,
                 as_worker=False, id=None, update_pipe_data=False):

        self.component = component
        self.pipeline = pipeline

        self.config = config

        self.update_pipe_data = update_pipe_data

        self._context_list = []

        self.as_worker = as_worker

        pre_id = id or name
        self.id = "%s:%s" % (str(pre_id),
                             unique_id_session.reserve_id_by_name(pre_id))
        self.name = name
        self.stepist_id = None

    def add_context(self, p_component, transformation):
        self._context_list.append(
            pipeline_context.ComponentContext(p_component, transformation)
        )

    def validate_output_data(self, data):
        output_data = dict()
        data_to_transform = dict()

        # remove all data related to current component
        for key, value in data.items():
            # skip values if it's not "subscribe" statement
            related_to_current_component = pipeline_context\
                .belongs_to_component(self, key)

            # skip data which was address to current component but it was
            # "apply" type component
            if related_to_current_component and not self.update_pipe_data:
                continue

            if not pipeline_context.is_context_key(key):
                if not self._context_list:
                    output_data[key] = value

                # collect data for transformation
                data_to_transform[key] = value
            else:
                # in case if data related to current component
                # but it should update global pipeline data, let's reassign
                # it to next component
                if related_to_current_component and self.update_pipe_data:
                    unassign_key = pipeline_context.unassign_data(self, key)
                    data_to_transform[unassign_key] = value
                else:
                    output_data[key] = value

        # in case if component has context list (has next components)
        # let's reassign data
        if self._context_list:
            # apply transformation and assign context labels
            for context in self._context_list:
                data = context.transformation(data_to_transform)
                output_data.update(context.assign_labels(data))
        else:
            # in case if current component is last one
            # and self.udpate_pipe_data, let's keep data
            if self.update_pipe_data:
                output_data.update(data_to_transform)

        return output_data

    def validate_input_data(self, data):
        input_data = dict()
        for key, value in data.items():
            if pipeline_context.belongs_to_component(self, key):
                new_key = pipeline_context.unassign_data(self, key)
                input_data[new_key] = value

        return input_data


class PipelineFlow(PipelineComponent):

    def __init__(self, pipeline, component, **kwargs):
        component.compile(pipeline)
        PipelineComponent.__init__(self,
                                   pipeline=pipeline,
                                   component=component,
                                   **kwargs)

    def __call__(self, **kwargs):
        component_data = self.validate_input_data(kwargs)
        flow_data = validate_handler_data(self.component, component_data)

        try:
            result = self.component(**flow_data)
        except StopPipelineFlag:
            raise StopFlowFlag()

        # if self.update_pipe_data:
        #     result_data = dict(**kwargs, **result)
        # else:
        #     result_data = result
        output = self.validate_output_data(dict(**kwargs, **result))
        return output


class PipelineFlowProducer(PipelineComponent):
    def __init__(self, pipeline, component, **kwargs):
        component.compile(pipeline)
        PipelineComponent.__init__(self,
                                   pipeline=pipeline,
                                   component=component,
                                   **kwargs)

    def __call__(self, **kwargs):
        component_data = self.validate_input_data(kwargs)

        flow_kwargs = validate_handler_data(self.component, component_data)

        try:
            result = self.component(**flow_kwargs)
        except StopPipelineFlag:
            raise StopFlowFlag()

        if isinstance(result, types.GeneratorType):
            for row_data in result:
                if self.update_pipe_data:
                    result_data = dict(**kwargs, **row_data)
                else:
                    result_data = row_data

                yield self.validate_output_data(result_data)
        else:
            raise RuntimeError("Flow producer should be a generator")


class PipelineFunction(PipelineComponent):
    def __call__(self, **kwargs):
        component_data = self.validate_input_data(kwargs)
        flow_data = validate_handler_data(self.component, component_data)
        result = self.component(**flow_data)

        if self.update_pipe_data:
            result_data = dict(**kwargs, **result)
        else:
            result_data = result

        return self.validate_output_data(result_data)


class PipelineOutput(PipelineComponent):
    def __call__(self, **kwargs):
        component_data = self.validate_input_data(kwargs)

        call_next_step(component_data, self.component.get_stepist_step())
        return self.validate_output_data(kwargs)


class PipelineInVainComponent(PipelineComponent):
    def __init__(self, pipeline, name="InVain"):
        PipelineComponent.__init__(self,
                                   pipeline=pipeline,
                                   component=None,
                                   config=None,
                                   as_worker=False,
                                   name=name)

    def __call__(self, **data):
        data.update(self.validate_input_data(data))
        data = self.validate_output_data(data)
        return data



from stairs.core.session import project_session

from stairs.core.output.output_model import Output


class StandAloneConsumer(Output):

    def __init__(self, *args, **kwargs):
        Output.__init__(self, *args, **kwargs, as_daemon=True)

    def __call__(self, *args, **kwargs):
        return self.handler(*args, **kwargs)

    def run_worker(self):
        stairs_project = project_session.get_project()
        worker_engine = stairs_project.stepist_app.worker_engine

        worker_engine.process(self.step)





class SimpleQueueStats:

    def __init__(self, name, amount_of_jobs, amount_of_time, jobs_processed):
        self.name = name
        self.amount_of_jobs = amount_of_jobs
        self.jobs_processed = jobs_processed
        self.amount_of_time = amount_of_time

    def get_jobs_per_sec(self):
        return self.jobs_processed / self.amount_of_time

    def to_dict(self):
        return dict(
            name=self.name,
            amount_of_jobs=self.amount_of_jobs,
            jobs_per_sec=self.get_jobs_per_sec(),
        )

    def print_info(self):
        print("Queue: %s" % self.name)
        print("Amount of jobs: %s" % self.amount_of_jobs)
        jobs_per_sec = self.get_jobs_per_sec()
        if jobs_per_sec > 0:
            print("Queue growing by %s tasks per/sec" % jobs_per_sec)
        else:
            print("Queue decreasing by %s tasks per/sec" % jobs_per_sec)


class MonitorQueueStats(SimpleQueueStats):

    def __init__(self, new_jobs, *args, **kwargs):
        self.new_jobs = new_jobs

        SimpleQueueStats.__init__(self, *args, **kwargs)

    def print_info(self):
        print("Queue: %s" % self.name)
        print("Amount of jobs: %s" % self.amount_of_jobs)

        print("New jobs per/sec: %s" % (self.new_jobs/self.amount_of_time))
        print("Jobs processed per/sec: %s" % (self.jobs_processed/self.amount_of_time))



import stepist

from stepist.flow import session
from stepist.flow.steps.next_step import call_next_step

from stairs.core.producer.adapter import BaseProducerAdapter


class IterAdapter(BaseProducerAdapter):

    # db cursor
    handler = None

    app_input = None

    def __init__(self, app, handler, app_inputs):
        self.app = app
        self.app_inputs = app_inputs or []
        self.handler = handler

    def init_session(self):
        """
        Init generator session
        """
        pass

    def process(self):
        try:
            for row in self.handler():
                with session.change_flow_ctx({}, {}):
                    if not isinstance(row, dict):
                        row = {'row': row}

                    for input in self.app_inputs:
                        call_next_step(row, input.step)

        except Exception:
            raise



from stairs.core.worker.pipeline_objects import PipelineComponent
from stairs.core.worker.worker import Worker

from stairs.core.output.standalone import StandAloneConsumer

PIPELINES_STEPS_TO_RUN = [
    PipelineComponent,
    Worker
]

STANDALONE_CONSUMERS_TO_RUN = [
    StandAloneConsumer
]




from . import local


def storage():
    if not hasattr(local, 'unique_ids'):
        local.unique_ids = dict(
            names=dict()
        )
    return local.unique_ids


def reserve_id_by_name(name):
    if name in storage()['names']:
        id = storage()['names'][name] + 1
    else:
        id = 0

    set_current_id_by_name(name, id)
    return id


def set_current_id_by_name(name, id):
    storage()['names'][name] = id



from stairs.core.output.output_model import Output


class Savepoint(Output):

    def __init__(self):
        Output.__init__(self, self.write, as_daemon=True)

    def read(self):
        raise NotImplementedError()

    def write(self):
        raise NotImplementedError()
